# Copyright Red Hat
#
# snapm/manager/plugins/__init__.py - Snapshot Manager plugins
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Snapshot manager plugin interface.
"""
from ._plugin import *  # noqa: F401, F403
from ._plugin import __all__  # noqa: F401
