# Copyright Red Hat
#
# snapm/__init__.py - Snapshot Manager package initialisation
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Snapm top-level package.
"""
from ._snapm import *  # noqa: F401, F403
from ._snapm import __all__  # noqa: F401

__version__ = "0.4.1"
